/*
 * Copyright (C) 2002 by Jirka Kosina <jikos@jikos.cz>
 * This program is free software under GPL.
 *  See ../COPYING for detailed info
 */


#include "japach.h"

char old_document_root[MAXLINE];
FILE *old_logfile;

 /*
  * tries to write all nbytes chars in loop (of *ptr)
  */
int writen(int fd, char *ptr, int nbytes)
{
	int nleft, nwritten;

	nleft = nbytes;
	while (nleft > 0) {
		nwritten = write(fd, ptr, nleft);
		if (nwritten <= 0)
			return (nwritten);

		nleft -= nwritten;
		ptr += nwritten;
	}
	return (nbytes - nleft);
}

 /* 
  * readss line of maximum length maxlen
  */
int readline(int fd, char *ptr, int maxlen)
{
	int n, rc;
	char c;

	for (n = 1; n < maxlen; n++) {
		if ((rc = read(fd, &c, 1)) == 1) {
			*ptr++ = c;
			if (c == '\n')
				break;
		} else if (rc == 0) {
			if (n == 1)
				return (0);	/* EOF, no data read */
			else
				break;	/* EOF, some data read */
		} else
			return (-1);	/* error */
	}

	*ptr = 0;
	return (n);
}

 /*
  * writes a line and doesn't care for result
  */

writestring(int fd, char *s1)
{
	int n;

	n = writen(fd, s1, strlen(s1));
	return (n);
}

  /*
   * writes default headers (FIXME: cgi scripts should
   * write it's own content-type kludge)
   */

int writedefaulthdr(int fd)
{
	char tmp[MAXLINE], tmp1[MAXLINE], *tmp2;
	int i;
	struct stat buf;

	if (!strlen(client.http_version))
		return;
	fstat(client.getfile_desc, &buf);
	strcpy(tmp1, "Date: ");
	get_rfc822_date(tmp);
	strncat(tmp, "\r\n", MAXLINE - strlen(tmp));
	strncat(tmp1, tmp, MAXLINE - strlen(tmp1));

	writestring(fd, tmp1);
	writestring(fd, "Server: JAPACH/");
	writestring(fd, VERSION);
	writestring(fd, "\r\n");

	if (client.reply != 304) {	/* a file will be output */
		writestring(fd, "Accept-Ranges: bytes\r\n");
		writestring(fd, "Connection: close\r\n");
	}

	if (client.cgi_state == CGI_OK)
		return 0;
	
	if ((!strip_extension(client.getfile))
	    || (client.reply == 404)
	    || (client.reply == 400)
	    || (client.reply == 417)
	    || (client.reply == 403)
	    || (client.reply == 301)
	    || (client.reply == 501)) {	/* no file is going to output */
		strcpy(tmp1, "Last-Modified: ");
		strncat(tmp1, tmp, MAXLINE - strlen(tmp1));
		writestring(fd, tmp1);

		if (!strip_extension(client.getfile))
			writestring(fd, "Content-Type: text/plain\r\n\r\n");
		else
			writestring(fd, "Content-Type: text/html\r\n\r\n");
	}

	else if (client.reply != 304) {
		int i;

		tmp2 = (char *) malloc(MAXLINE * sizeof(char));
		convert_to_rfc822_date(buf.st_mtime, tmp2);
		strcpy(tmp, "Last-Modified: ");
		strncat(tmp, tmp2, MAXLINE - strlen(tmp) - 2);
		strcat(tmp, "\r\n");

		free(tmp2);
		writestring(fd, tmp);
		/* going to find content-type according to extension, and output
		 * apropriate content-type
		 */
		strcpy(tmp, strip_extension(client.getfile));
		tmp2 = find_content_type(tmp);
		strcpy(tmp1, "Content-Type: ");
		strncat(tmp1, tmp2, MAXLINE - strlen(tmp1) - 3);
		/* strcat(tmp1, "\r\n\n"); */

		writestring(fd, tmp1);
	}

}

 /*
  * generates http reply to socked, sets client.reply properly, opens
  * local file (assigns descriptor)
  *          - cgi started 17/02/2002
  */

int genreply(int fd)
{
	char tmp[MAXLINE], tmp1[MAXLINE], *tmp2, *found_s;
	char link_getfile[MAXLINE];
	char text[MAXLINE];
	char username[MAXLINE], tosubst[MAXLINE], userpath[MAXLINE];
	int i;
	int found = 0;
	struct stat buf;
	struct stat lbuf;
	struct virtuals *pom;
	struct passwd *psw;

	if (!strcasecmp(client.method, "GET")) {	/* GET-specific actions */
		/* remove part following ?
		 * TODO: RFC?
		 */
		if (client.cgi_get_query_string = strchr(client.req_url, '?')) {
			client.cgi_get_query_string[0] = '\0';
			client.cgi_get_query_string++;
		}

	} else if (!strcasecmp(client.method, "HEAD")) {
	} else {
		/* unknown method */
		if (strlen(client.http_version)) {
			writestring(fd, "HTTP/1.1 501 Method Not Implemented\r\n");
			writestring(fd, "Allow: GET\r\n");
		}
		logrequest(log.logfile, 501, strlen(HTTP_ERR_501));
		writedefaulthdr(fd);
		return;
	}
	if (strlen(client.expect)) {
		client.reply = 417;
		if (strlen(client.http_version))
			writestring(fd, "HTTP/1.1 417 Expectation failed\r\n");
		writedefaulthdr(fd);
		return;
	}
	if (client.req_url[0] != '/') {
		client.reply = 400;
		if (strlen(client.http_version))
			writestring(fd, "HTTP/1.1 400 Bad Request\r\n");
		writedefaulthdr(fd);
		return;
	}

	client.reply = -1;

	if ((!strcasecmp(client.http_version, "HTTP/1.1"))
	    || (!strcasecmp(client.http_version, "HTTP/1.0"))) {

		/* save default settings, after serving a request set old values.
		 * no use now, but in the future this may be rewritten to be able
		 * to serve more than one request during one connection.
		 * so this is neccessary to have the right defaults for the next
		 * request
		 */
		strcpy(old_document_root, conf.document_root);
		old_logfile = log.logfile;

		/* find in conf->virtual_hosts name of virtual server, copy to
		 * document_root proper conf->virtual_hosts->documentroot
		 */
		pom = conf.virtual_hosts->next;
		while (pom && !found) {
			if (!strcasecmp(client.requested_host, pom->name)) {
				found = 1;
				strncpy(conf.document_root, pom->documentroot, MAXLINE);
				log.logfile = pom->logfile;
			} else
				pom = pom->next;
		}
	}
	/* all changing of URL (filename, respectively) is done here
	 * decoding of special %ab characters, ~ expansion, ...
	 */
	tmp2 = decode_hexa_octets(client.req_url);
	strcpy(client.req_url, tmp2);
	free(tmp2);

	if (client.req_url[1] == '~') {
		i = 2;
		while ((i < strlen(client.req_url)) && (client.req_url[i] != '/')) {
			username[i - 2] = client.req_url[i];
			username[i - 1] = '\0';
			i++;
		}
		found = i;
		while (i < strlen(client.req_url)) {
			tosubst[i - found] = client.req_url[i];
			tosubst[i - found + 1] = '\0';
			i++;
		}
		if (psw = getpwnam(username)) {
			strcpy(conf.document_root, psw->pw_dir);
			strcat(conf.document_root, "/");
			strncat(conf.document_root, conf.userdir, MAXLINE);
			if (client.req_url[i] == '/')
				strcat(conf.document_root, "/");
		}
	}
	strcpy(client.getfile, conf.document_root);
	if (client.req_url[1] != '~')
		strncat(client.getfile, client.req_url, MAXLINE);
	else
		strncat(client.getfile, tosubst, MAXLINE);
	tmp2 = deldupl('/', client.getfile);
	strcpy(client.getfile, tmp2);


	if (is_in_cage(client.getfile)) {
		client.getfile_desc = open(client.getfile, O_RDONLY);
		if (client.getfile_desc > 0) {
			fstat(client.getfile_desc, &buf);
			if (S_ISDIR(buf.st_mode)) {
				if (client.getfile[strlen(client.getfile) - 1] != '/')
					client.reply = 301;
				else {
					if (strlen(client.getfile) < MAXLINE - 12) {
						strncat(client.getfile, "/index.html",
							MAXLINE - strlen(client.getfile));
						close(client.getfile_desc);
						client.getfile_desc =
						    open(client.getfile, O_RDONLY);
					} else
						client.reply = 404;
				}
			}
			lstat(client.getfile, &lbuf);

			/* let's do some symlink handling */

			if ((S_ISLNK(lbuf.st_mode)))
				getlink(client.getfile, link_getfile);
			if ((S_ISLNK(lbuf.st_mode)
			     && (conf.follow_symlinks == DO_NOT_FOLLOW_SYMLINKS))
			    || (S_ISLNK(lbuf.st_mode)
				&& (conf.follow_symlinks == FOLLOW_SYMLINKS_INTO_CAGE)
				&& (!is_in_cage(link_getfile)))) {
				client.reply = 403;
				if (strlen(client.http_version)) {
					writestring(fd, "HTTP/1.1 403 Forbidden\r\n");
					writedefaulthdr(fd);
				}
				logrequest(log.logfile, 403, strlen(HTTP_ERR_403));
			}
			if ((client.getfile_desc > 0) && (client.reply != 301)
			    && (client.reply != 403) && (client.reply != 404)) {
				if (is_in_cage(client.getfile)) {
					if (strlen(client.if_mod_since)) {	/* if-modified-since */
						tmp2 = (char *) malloc(MAXLINE * sizeof(char));
						convert_to_rfc822_date(buf.st_mtime, tmp2);
						if (compare_rfc822_dates(tmp2, client.if_mod_since)
						    == -1)
							client.reply = 304;
						free(tmp2);
					}
					if (strlen(client.if_unmod_since)) {	/* If-Unmodified-Since */
						tmp2 = (char *) malloc(MAXLINE * sizeof(char));
						convert_to_rfc822_date(buf.st_mtime, tmp2);
						if (compare_rfc822_dates
						    (tmp2, client.if_unmod_since) == 1)
							client.reply = 412;
						free(tmp2);
					}
					if ((client.reply != 304) && (client.reply != 412)) {
						/* the file is ok, now we need to check if it has to (and could)
						 * be run (CGI). If so, then 200, otherwise 403 */
						client.cgi_state = if_cgi_then_ok(client.getfile);
						if ((client.cgi_state == CGI_NOCGI)
						    || (client.cgi_state == CGI_OK)) {
							client.reply = 200;
							if (strlen(client.http_version)) {
								writestring(fd,
									    "HTTP/1.1 200 OK\r\n");
								writedefaulthdr(fd);
							}

							logrequest(log.logfile, 200, buf.st_size);
						} else {	// cgi_err
							client.reply = 403;
							if (strlen(client.http_version)) {
								writestring(fd,
									    "HTTP/1.1 403 Forbidden\r\n");
								writedefaulthdr(fd);
							}
							logrequest(log.logfile, 403, buf.st_size);
						}
					} else {
						if (client.reply == 304) {
							writestring(fd,
								    "HTTP/1.1 304 Not Modified\r\n");
							logrequest(log.logfile, 304, 0);
							writedefaulthdr(fd);
						} else if (client.reply == 412) {
							writestring(fd,
								    "HTTP/1.1 412 Precondition Failed\r\n");
							logrequest(log.logfile, 412, 0);
							writedefaulthdr(fd);

						}
					}
				} else {
					client.reply = 403;
					if (strlen(client.http_version)) {
						writestring(fd, "HTTP/1.1 403 Forbidden\r\n");
						writedefaulthdr(fd);
					}
					logrequest(log.logfile, 403, strlen(HTTP_ERR_403));
				}
			}
		}
	} else {		/* cage-membership test failed */
		client.reply = 403;
		if (strlen(client.http_version)) {
			writestring(fd, "HTTP/1.1 403 Forbidden\r\n");
			writedefaulthdr(fd);
		}
		logrequest(log.logfile, 403, strlen(HTTP_ERR_403));
	}
	if (client.reply == 301) {
		if (strlen(client.http_version)) {
			writestring(fd, "HTTP/1.1 301 Moved Permanently\r\n");
			writestring(fd, "Location: ");

			strcpy(text, "http://");
			strncat(text, client.requested_host, MAXLINE - strlen(text));
			strncat(text, client.req_url, MAXLINE - strlen(text));
			strcat(text, "/\r\n");

			writestring(fd, text);
			writedefaulthdr(fd);
		}
		logrequest(log.logfile, 301,
			   strlen(HTTP_ERR_301_P1) + strlen(HTTP_ERR_301_P2) + 2 * strlen(text) +
			   14);

	}
	if (((client.getfile_desc < 0)) && ((errno != EACCES) || (client.reply == 404))) {	/* we can have 404 from earlier */
		client.reply = 404;
		if (client.getfile > 0)
			close(client.getfile_desc);
		if (strlen(client.http_version)) {
			writestring(fd, "HTTP/1.1 404 Not Found\r\n");
			writedefaulthdr(fd);
		}
		logrequest(log.logfile, 404, strlen(HTTP_ERR_403));
	}
	if ((client.getfile_desc < 0) && (errno == EACCES)) {
		client.reply = 403;
		if (strlen(client.http_version)) {
			writestring(fd, "HTTP/1.1 403 Forbidden\r\n");
			writedefaulthdr(fd);
		}
		logrequest(log.logfile, 403, strlen(HTTP_ERR_403));
	}
}

 /*
  * read headers and process content of client-provided data
  * TODO: some macros would do a good job in following
  * boring & nasty code
  */

int read_headers(int sockfd)
{
	int n, i;
	char line[MAXLINE];
	int numlines = 0;

	numlines++;
	n = readline(sockfd, line, MAXLINE);
	line[strlen(line) - 2] = '\0';

	if (numlines == 1)
		strcpy(client.first_line, line);

	/* parse method */
	i = 0;
	while ((i < strlen(line)) && (line[i] != ' ')) {
		client.method[i] = line[i];
		i++;
	}
	client.method[i] = '\0';

	/* URL */
	i = strlen(client.method) + 1;
	while ((line[i] != ' ') && (i <= strlen(line))) {
		client.req_url[i - (strlen(client.method) + 1)] = line[i];
		i++;
	}
	client.req_url[i - strlen(client.method)] = '\0';

	/* HTTP protocl version */
	while (line[i] == ' ')
		i++;
	n = i;
	while ((line[i] != ' ') && (i <= strlen(line))) {
		client.http_version[i - n] = line[i];
		i++;
	}
	client.http_version[i - (n - 1)] = '\0';

	/* if the request is in old HTTP version, don't wait for 
	 * more headers & clear client.http_version
	 */
	if ((!strlen(client.http_version))) {
		strcpy(line, "");
		strcpy(client.http_version, "");
	}

	/* more and more headers */
	strcpy(client.browser, "");
	strcpy(client.requested_host, conf.server_name);
	strcpy(client.if_mod_since, "");

	while ((strlen(line) != 0) && (n != 0)) {
		n = readline(sockfd, line, MAXLINE);	/* wait for and empty line */
		line[strlen(line) - 2] = '\0';
		if (strncasecmp(line, "User-agent", 10) == 0) {
			int i = 12;
			while ((isspace(line[i])) && (i <= strlen(line)))
				i++;
			strncpy(client.browser, line + i, MAXLINE);
		}
		if (strncasecmp(line, "host:", 5) == 0) {
			int i = 6;
			while ((isspace(line[i])) && (i <= strlen(line)))
				i++;
			strncpy(client.requested_host, line + i, MAXLINE);
			i = 0;

			while (isalnum(client.requested_host[i]) || client.requested_host[i] == '.' ||
					client.requested_host[i] == '-')
				i++;
			if (i)
				client.requested_host[i] = '\0';
		}
		if (strncasecmp(line, "If-Modified-since:", 18) == 0) {
			int i = 19;
			while ((isspace(line[i])) && (i <= strlen(line)))
				i++;
			strncpy(client.if_mod_since, line + i, MAXLINE);
		}
		if (strncasecmp(line, "If-Unmodified-since:", 20) == 0) {
			int i = 21;
			while ((isspace(line[i])) && (i <= strlen(line)))
				i++;
			strncpy(client.if_unmod_since, line + i, MAXLINE);
		}
		if (strncasecmp(line, "Expect:", 7) == 0) {
			int i = 8;
			while ((isspace(line[i])) && (i <= strlen(line)))
				i++;
			strncpy(client.expect, line + i, MAXLINE);
		}
	}
	if (n == 0) {
	/* no problem here - no need to flush anything, and the OS
	 * will do what is needed not to leak descriptors
	 */
#ifdef HAVE_FCLOSEALL
		fcloseall();
#endif
		free_mem();
		exit();
	}
	return (0);
}

 /* routine called from freshly forked child from main
  * program loop in japach.c
  */
int process_request(int sockfd)
{
	char text[MAXLINE];
	void *data;

	for (;;) {

		read_headers(sockfd);

		/* here client.reply is to be generated, which is
		 * then tested
		 * also some url-rewrites are done, and local file
		 * descriptor is assigned
		 */

		genreply(sockfd);
		if (client.cgi_state != CGI_OK)
			writestring(sockfd, "\r\n\n");			
		if (!strcasecmp(client.method, "HEAD"))
			return;
		switch (client.reply) {
		case 200:{	// ok
				int i;
				signal(SIGCHLD, SIG_IGN);
				if (client.cgi_state == CGI_OK) {
					int cgi_pipe[2], bytes;
					char line[MAXLINE];
					pid_t pid;
					char **tmp;

					pipe(cgi_pipe);
					pid = fork();

					if (pid == 0) {
						close(cgi_pipe[0]);
						dup2(cgi_pipe[1], 1);	/* 1 = stdout */
						tmp = cgi_get_environment();
						execve(client.getfile, NULL, cgi_get_environment());
					} else {
						close(cgi_pipe[1]);
						write(sockfd, "BLABLE\n",7);
						while ((bytes =
							read(cgi_pipe[0], line, MAXLINE)) > 0) {
							if (bytes) write(sockfd, line, bytes);
							memset(line, '\0', sizeof(line));
						}
						close(cgi_pipe[0]);
						waitpid(pid, NULL, 0);
					}
				} else {
					data = (char *) malloc((sizeof(char)) * MAXLINE + 1);
					i = MAXLINE;
					while (i == MAXLINE)
						if ((i =
						     read(client.getfile_desc, data,
							  MAXLINE)) == MAXLINE)
							writen(sockfd, data, i);	/* rewrite file to network */
					if (i > 0)
						writen(sockfd, data, i);	/* negative i meands error from read() */
					close(client.getfile_desc);
					free(data);
				}
				/* put back original values (may have been changed
				 * by virtualhost)
				 */
				strcpy(conf.document_root, old_document_root);
				log.logfile = old_logfile;

				/* here is the point, where one than more requests during one connection
				 * is to be handled. maybe recursive call of process_request() is not
				 * the best idea, but should work ;)

				 process_request(sockfd);
				 */

				return;
			}
		case 301:{	/* moved permanently */
				strcpy(text, HTTP_ERR_301_P1);
				strcat(text, "http://");
				strncat(text, client.requested_host, MAXLINE - strlen(text));
				strncat(text, client.req_url, MAXLINE - strlen(text));
				strcat(text, "/");
				strcat(text, "\">");
				strcat(text, "http://");
				strncat(text, client.requested_host, MAXLINE - strlen(text));
				strncat(text, client.req_url, MAXLINE - strlen(text));
				strcat(text, "/");
				strcat(text, "</a>\r\n");
				strncat(text, HTTP_ERR_301_P2, MAXLINE - strlen(text));
				writestring(sockfd, text);
				return;
			}
		case 304:{	/* not modified */
				writestring(sockfd, "");
				return;
			}
		case 400:{	/* bad request */
				strcpy(text, HTTP_ERR_400);
				writestring(sockfd, text);
				return;
			}
		case 404:{	/* not found */
				strcpy(text, HTTP_ERR_404);
				writestring(sockfd, text);
				return;
			}
		case 403:{	/* forbidden */
				strcpy(text, HTTP_ERR_403);
				writestring(sockfd, text);
				return;
			}
		case 412:{	/* precondition failed */
				strcpy(text, HTTP_ERR_412);
				writestring(sockfd, text);
				return;
			}
		case 417:{	/* expectation failed */
				writestring(sockfd, "");
				return;
			}
		case 501:{	/* method not implemented */
				strcpy(text, HTTP_ERR_501);
				writestring(sockfd, text);
				return;
			}
		}
	}
}

 /* test if a file is located "in the cage".
  * cage is a struct read from conf file and
  * stored in conf.cage_dirs linked list
  */

int is_in_cage(char *s)
{
	struct cage *pom;
	char *pom1;
	int found = 0;

	pom1 = (char *) malloc(sizeof(char) * MAXLINE);
	delddot(s, pom1);
	s = pom1;
	pom = conf.cage_dirs;
	while ((!found) && (pom->next)) {
		if (strstr(s, pom->next->name) == s)
			return (1);
		else
			pom = pom->next;
	}
	return (0);
}
