/*
 * Copyright (C) 2002 by Jirka Kosina <jikos@jikos.cz>
 * This program is free software under GPL.
 * See ../COPYING for detailed info
 */

#include "japach.h"

int childs = 0;			/* count of forked children */
int first_cycle = 1;

sig_atomic_t terminate, reload;

void flush_virtual_logs()
{
	struct virtuals *p;

	p = conf.virtual_hosts;
	while (p){
		fflush(p->logfile);
		p = p->next;
	}
}

 /* TODO: some macros would be nice here */
void free_mem()
{
	/* -- struct conf -- */
	struct c_types *p1, *p11;
	struct cage *p2, *p22;
	struct virtuals *p3, *p33;
	struct cgidirs *p4, *p44;
	p1 = conf.content_types;
	p11 = p1->next;

	p2 = conf.cage_dirs;
	p22 = p2->next;

	p3 = conf.virtual_hosts;
	p33 = p3->next;

	p4 = conf.cgi_dirs;
	p44 = p4->next;

	while (p1) {
		p11 = p1->next;
		free(p1);
		p1 = p11;
	}

	while (p2) {
		p22 = p2->next;
		free(p2);
		p2 = p22;
	}

	while (p3) {
		p33 = p3->next;
		free(p3);
		p3 = p33;
	}

	while (p4) {
		p44 = p4->next;
		free(p4);
		p4 = p44;	/* no need to close file */
	}

}

void f_SIGCHLD()
{
	union wait wstatus;
	pid_t exit_pid;
	
	while ((exit_pid = wait3(&wstatus, WNOHANG, NULL)) > 0) { 
		if (exit_pid != -1)
/*			fprintf(stderr,"Child with pid %d exited, reducing to %d\n", 
					exit_pid, childs - 1);*/
		childs--;
	}
}

void f_SIGTERM()
{
	logwrite(log.logfile, "caught SIGTERM\n");
	terminate = 1;
}

void f_SIGINT()
{
	logwrite(log.logfile, "caught SIGINT\n");
	terminate = 1;
}

void f_SIGHUP()
{
	logwrite(log.logfile, "caught SIGHUP, resetting\n");
	reload = 1;
}

main(int argc, char *argv[])
{
	int sockfd, newsockfd, clilen, childpid;
	struct sockaddr_in cli_addr, serv_addr;
	struct hostent *hostentry;
	FILE *conf_file;
	struct sigaction *s1, *s2;
	int i, opt = 1;
	pid_t pidstart;
	struct cage *pom;


	/* config */
	strcpy(conf_file_name, "japach.conf");
	for (i = 1; i < argc; i++) {
		if ((!strcmp(argv[i], "-c")) && (i < argc - 1)) {
			strcpy(conf_file_name, argv[i + 1]);
		}
	}

	/* start daemon
	 */
	if ((pidstart = fork()) < 0) {
		fprintf(stderr, "master fork error.\n");
		return (-1);
	} else if (pidstart > 0) {
		printf("master fork ok, child daemon (pid %d) created\n",
		       pidstart);
		return (0);
	}

	terminate = 0;
	while (!terminate) {

		conf_file = fopen(conf_file_name, "r");
		if (!conf_file) {
			fprintf(stderr, "conf file not found\n");
			return (-1);
		}
		if (!first_cycle) {
			free_mem();
			first_cycle = 0;
		}
		parse_config(conf_file);
		fclose(conf_file);


		if (!(log.logfile = fopen(conf.master_log, "a"))) {
			fprintf(stderr, "can't open %s\n",
				conf.master_log);
			free_mem();
			exit();
		}
		logwrite(log.logfile, "Server starting\n");

		s1 = (struct sigaction *) malloc(sizeof(struct sigaction));
		s2 = (struct sigaction *) malloc(sizeof(struct sigaction));
		s1->sa_handler = f_SIGHUP;
		sigaction(SIGHUP, s1, s2);
		s1->sa_handler = f_SIGTERM;
		sigaction(SIGTERM, s1, s2);
		s1->sa_handler = f_SIGINT;
		sigaction(SIGINT, s1, s2);
		s1->sa_handler = f_SIGCHLD;
		s1->sa_flags = SA_RESTART;
		sigaction(SIGCHLD, s1, s2);
		free(s1);
		free(s2);

		/* socket ops */
		if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
			logwrite(log.logfile,
				 "can't open stream socket.\n");
			f_SIGTERM();
		}

		if (setsockopt
		    (sockfd, SOL_SOCKET, SO_REUSEADDR, &opt,
		     sizeof(opt)) != 0) {
			fprintf(stderr, "error setting socket options.\n");
			logwrite(log.logfile,
				 "error setting socket options.\n");
			f_SIGTERM();
		}
		bzero((char *) &serv_addr, sizeof(serv_addr));
		serv_addr.sin_family = AF_INET;
		serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
		serv_addr.sin_port = htons(conf.serv_tcp_port);

		if (bind
		    (sockfd, (struct sockaddr *) &serv_addr,
		     sizeof(serv_addr)) < 0) {
			fprintf(stderr, "error binding address.\n");
			logwrite(log.logfile, "error binding address.\n");
			close(sockfd);
			f_SIGTERM();
		}

		if (listen(sockfd, 5) != 0) {
			fprintf(stderr, "error calling listen()\n");
			logwrite(log.logfile, "error calling listen()\n");
		}


		reload = 0;
		while (!terminate && !reload) {
			while (childs >= conf.max_childs)
				pause();	/* wait until signal occurs.
						 * we are waiting for SIGCHLD to decrement
						 * childs variable
						 */

			clilen = sizeof(cli_addr);
			if (!
			    (newsockfd =
			     accept(sockfd, (struct sockaddr *) &cli_addr,
				    &clilen))) {
				fprintf(stderr, "accept error\n");
				exit(0);
			}

			if ((terminate) || (reload))
				continue;

			/* reverse resolving */
			if (NULL !=
			    (hostentry =
			     gethostbyaddr((char *) &cli_addr.sin_addr,
					   sizeof(cli_addr.sin_addr),
					   AF_INET)))
				client.host_name = hostentry->h_name;
			else
				client.host_name =
				    inet_ntoa(cli_addr.sin_addr);
			client.ip_address = inet_ntoa(cli_addr.sin_addr);

			/* k* child, which will serve the request */
			if ((childpid = fork()) < 0)
				printf("server: fork error\n");
			else if (childpid == 0) {
				struct sigaction *s1, *s2;

				setgid(conf.gid);
				setuid(conf.uid);
				close(sockfd);

				s1 = (struct sigaction *) malloc(sizeof(struct sigaction));
				s2 = (struct sigaction *) malloc(sizeof(struct sigaction));
				s1->sa_handler = SIG_IGN;
				sigaction(SIGHUP, s1, s2);
				free(s1);
				free(s2);
				process_request(newsockfd);	/* GO! */
				/*fprintf(stderr,"Processing (%d) done, exiting...\n",getpid());*/
				close(newsockfd);
				exit(0);
			} else if (childpid > 0){
				childs++;	/* increment count of forked childs in parent process */
				/*fprintf(stderr, "New child with pid %d created, incrementing to %d\n",
						childpid, childs);*/
			}
			close(newsockfd);
		}
		close(sockfd);
		fclose(log.logfile);
	}			/* end of main loop */
	logwrite(log.logfile, "END\n");
#ifdef HAVE_FCLOSEALL
	fcloseall();
#else
	fflush(log.logfile);
	flush_virtual_logs();
#endif
	free_mem();
	free(s1);
	exit();
}
