/*
 * Copyright (C) 2002 by Jirka Kosina <jikos@jikos.cz>
 * This program is free software under GPL.
 * See ../COPYING for detailed info
 */

#include    <sys/types.h>
#include    <sys/stat.h>
#include    <unistd.h>
#include    <fcntl.h>
#include    <sys/socket.h>
#include    <string.h>
#include    <netdb.h>
#include    <time.h>
#include    <stdio.h>
#include    <netinet/in.h>
#include    <arpa/inet.h>
#include    <signal.h>
#include    <errno.h>
#include    <sys/wait.h>
#include    <setjmp.h>
#include    <pwd.h>



#define VERSION "0.4.8"
#define FULL_SERVER_SOFTWARE_NAME "japach/"VERSION

 /* definice max. velikosti promennych */
#define MAXSEG 2048
#define MAXLINE 2048
#define MAX_ENVIRON_VARS 30

 /* definice chybovych hlasek */
#define HTTP_ERR_301_P1 "\n\n<html><body>301 - Moved Permanently to <a href=\""
#define HTTP_ERR_301_P2 "</body></html>\n\n"

#define HTTP_ERR_400 "\n\n<html><body>400 - Bad request</body></html>\n\n"
#define HTTP_ERR_404 "\n\n<html><body>404 - Not found. Sorry</body></html>\n\n"
#define HTTP_ERR_403 "\n\n<html><body>403 - This file/path is forbidden.</body></html>\n\n"
#define HTTP_ERR_412 "\n\n<html><body>412 - Precondition failed.</body></html>\n\n"

#define HTTP_ERR_501 "\n\n<html><body>501 - Method Not Implemented.</body></html>\n\n"

 /* definice konstant */
#define DO_NOT_FOLLOW_SYMLINKS 0
#define FOLLOW_SYMLINKS_INTO_CAGE 1
#define FOLLOW_SYMLINKS 2

#define CGI_OK 1		// jedna se o cgi skript, a je ok
#define CGI_ERR 0		// jedna se o cgi skript, a neni ok
#define CGI_NOCGI 2		// nejedna se o cgi skript

 /* globalni promenne */
char conf_file_name[MAXLINE];

 /* struktura pro udrzovani content-typu */
struct c_types {
	char extension[MAXLINE];
	char content_type[MAXLINE];
	struct c_types *next;
};

 /* struktura pro udrzovani omezujicih adresaru */
struct cage {
	char name[MAXLINE];
	struct cage *next;
};

 /* struktura pro udrzovani adresaru, ve kterych je mozne spoustet CGI skripty */
struct cgidirs {
	char name[MAXLINE];
	struct cgidirs *next;
};

struct virtuals {
	char name[MAXLINE];
	char documentroot[MAXLINE];
	char log[MAXLINE];
	FILE *logfile;
	struct virtuals *next;
};

 /* konfiguracni struktura */
struct {
	char document_root[MAXLINE];
	char master_log[MAXLINE];
	char server_name[MAXLINE];
	char userdir[MAXLINE];
	unsigned int serv_tcp_port;
	unsigned int uid;
	unsigned int gid;
	unsigned int max_childs;
	unsigned int follow_symlinks;
	struct c_types *content_types;
	struct cage *cage_dirs;
	struct cgidirs *cgi_dirs;
	struct virtuals *virtual_hosts;
} conf;

 /* infomace o aktualnim klientovi */
struct {
	char *ip_address;
	char *host_name;
	char first_line[MAXLINE];

	char browser[MAXLINE];
	char req_url[MAXLINE];
	char *cgi_get_query_string;	/* v pripade cgi metodou GET prislusny query string */
	char http_version[MAXLINE];
	char method[MAXLINE];
	char requested_host[MAXLINE];
	char if_mod_since[MAXLINE];
	char if_unmod_since[MAXLINE];
	char expect[MAXLINE];

	int reply;		/* odpoved ktera se mu bude generovat */
	char getfile[MAXLINE];	/* realna cesta k predavanemu souboru */
	int getfile_desc;	/* deskriptor pro cteny soubor */

	unsigned int cgi_state;	/* jak je na tom tento pozadavek s CGI */
} client;

 /* logovani */
struct {
	FILE *logfile;
	char message[MAXLINE];
} log;

 /* -----------------deklarace funkci---------------------- */
 /* text.c */
void convert_to_rfc822_date(time_t curtime, char *buffer);
void get_rfc822_date(char *buffer);
char *strip_extension(char *buffer);
void prase_config(FILE * f);
int empty_string(char *s);
char *find_content_type(char *ext);
char *subst(char *a, char *b, char *c);
char *deldupl(char a, char *b);
char *decode_hexa_octets(char *s1);
int delddot(char *s, char *s2);
int get_number_of_month(char *tmp2);
int compare_rfc822_dates(char *date1, char *date2);

 /* communication.c */
int writen(int fd, char *ptr, int nbytes);
int readline(int fd, char *ptr, int maxlen);
int writestring(int fd, char *s1);
int genreply(int fd);
int read_headers(int sockfd);
int process_request(int sockfd);
int is_in_cage(char *s);

 /* cgi.c */
int logrequest(FILE * lf, int reply, int bytes);
int logwrite(FILE * lf, const char *s);

 /* sys.c */
void getlink(char *path, char *result);

 /* cgi.c */
int if_cgi_then_ok(char *fname);
char **cgi_get_environment();

