/*
 * Copyright (C) 2002 by Jirka Kosina <jikos@jikos.cz>
 * This program is free software under GPL.
 * See ../COPYING for detailed info
 */

#include "japach.h"

 /* converts time_t to rfc822 date */
void convert_to_rfc822_date(time_t curtime, char *buffer)
{
	struct tm *loctime;

	loctime = localtime(&curtime);
	strftime(buffer, 256, "%a, %d %b %Y %T %z", loctime);
}

void get_rfc822_date(char *buffer)
{
	time_t curtime;
	struct tm *loctime;

	curtime = time(NULL);
	loctime = localtime(&curtime);
	strftime(buffer, 256, "%a, %d %b %Y %T %z", loctime);
}

 /* strips  extension from /some/file/path.ext */
char *strip_extension(char *buffer)
{
	char *s1;
	int i;

	for (s1 = buffer + strlen(buffer); ((*s1 != '.') && (s1 > buffer));
	     s1--);
	s1++;
	if (s1 > buffer)
		return (s1);
	else
		return (0);
}

 /* simple */
int empty_string(char *s)
{
	int i = 0;

	while ((isspace(s[i])) && (i < strlen(s)))
		i++;
	if (i == strlen(s))
		return (1);
	else
		return (0);
}

/* parse config file */
/* this isn't indented in K&R style, because it will look ugly
 * (sorry guys ;) ).
 */

void parse_config(FILE * f)
{
	char *radka, delim, *confval, *keyword, *extension, *content_type,
	    *s1, *s2;
	char *name, *documentroot, *log;
	int i = 0, x = 0, cur_conf_line = 0;
	struct c_types *pom;
	struct cage *pom1;
	struct virtuals *pom2;
	struct cgidirs *pom3;

	/* let's set some defaults */
	conf.serv_tcp_port = 80;
	strcpy(conf.document_root, "/var/www/html/");
	strcpy(conf.master_log, "tcp_srv.log");
	strcpy(conf.server_name, "localhost");
	strcpy(conf.userdir, "html");
	conf.uid = 0;
	conf.gid = 0;
	conf.max_childs = 100;
	conf.content_types =
	    (struct c_types *) malloc(sizeof(struct c_types));
	conf.content_types->next = NULL;
	conf.cage_dirs = (struct cage *) malloc(sizeof(struct cage));
	conf.cage_dirs->next = NULL;
	conf.cgi_dirs = (struct cgidirs *) malloc(sizeof(struct cgidirs));
	conf.cgi_dirs->next = NULL;
	conf.virtual_hosts =
	    (struct virtuals *) malloc(sizeof(struct virtuals));
	conf.virtual_hosts->next = NULL;
	conf.follow_symlinks = DO_NOT_FOLLOW_SYMLINKS;

	radka = (char *) malloc(MAXLINE * sizeof(char));
	name = (char *) malloc(MAXLINE * sizeof(char));
	documentroot = (char *) malloc(MAXLINE * sizeof(char));
	log = (char *) malloc(MAXLINE * sizeof(char));
	keyword = (char *) malloc(MAXLINE * sizeof(char));
	confval = (char *) malloc(MAXLINE * sizeof(char));
	extension = (char *) malloc(MAXLINE * sizeof(char));
	content_type = (char *) malloc(MAXLINE * sizeof(char));


	while (!feof(f)) {
		cur_conf_line++;
		if ((fgets(radka, MAXLINE, f)) && (radka[0] != '#')
		    && (!empty_string(radka))) {
			i = 0;
			while ((!isspace(radka[i])) && (i < strlen(radka))) {
				keyword[i] = radka[i];
				i++;
			}
			keyword[i] = '\0';
			while ((isspace(radka[i])) && (i < strlen(radka)))
				i++;
			x = i;
			if (i < strlen(radka)) {
				while (i < strlen(radka)) {
					confval[i - x] = radka[i];
					i++;
				}
				confval[i - x - 1] = '\0';
			} else {
				fprintf(stderr,
					"keyword without value on line %d\n",
					cur_conf_line);
				exit();	/* FIXME: what about tidying up the mem? */
			}
			if (!strcasecmp(keyword, "port")) {
				conf.serv_tcp_port = atoi(confval);
			} else if (!strcasecmp(keyword, "documentroot")) {
				strcpy(conf.document_root, confval);
				if (confval[strlen(confval) - 1] != '/')
					strcat(conf.document_root, "/");
			} else if (!strcasecmp(keyword, "masterlog")) {
				strcpy(conf.master_log, confval);
			} else if (!strcasecmp(keyword, "uid")) {
				conf.uid = atoi(confval);
			} else if (!strcasecmp(keyword, "gid")) {
				conf.gid = atoi(confval);
			} else if (!strcasecmp(keyword, "maxchilds")) {
				conf.max_childs = atoi(confval);
			} else if (!strcasecmp(keyword, "symlinks")) {
				if (!strcasecmp
				    (confval, "follow_symlinks"))
					conf.follow_symlinks =
					    FOLLOW_SYMLINKS;
				else if (!strcasecmp
					 (confval,
					  "follow_symlinks_into_cage"))
					conf.follow_symlinks =
					    FOLLOW_SYMLINKS_INTO_CAGE;
			}	//jinak je default
			else if (!strcasecmp(keyword, "servername")) {
				strcpy(conf.server_name, confval);
			} else if (!strcasecmp(keyword, "userdir")) {
				strcpy(conf.userdir, confval);
			} else if (!strcasecmp(keyword, "cage")) {	//definice ze kterych adresaru nesmi ven
				pom1 = conf.cage_dirs;
				while (pom1->next)
					pom1 = pom1->next;
				pom1->next =
				    (struct cage *)
				    malloc(sizeof(struct cage));
				strcpy(pom1->next->name, confval);
				pom1->next->next = NULL;
			} else if (!strcasecmp(keyword, "cgidir")) {	/* paths to cgi executables */
				pom3 = conf.cgi_dirs;
				while (pom3->next)
					pom3 = pom3->next;
				pom3->next =
				    (struct cgidirs *)
				    malloc(sizeof(struct cgidirs));
				strcpy(pom3->next->name, confval);
				pom3->next->next = NULL;
			} else if (!strcasecmp(keyword, "type")) {	/* assigning content-types to extensions */
				pom = conf.content_types;
				while (pom->next)
					pom = pom->next;
				pom->next =
				    (struct c_types *)
				    malloc(sizeof(struct c_types));
				/* parsing content-type extension */
				i = 4;
				while (isspace(radka[i]))
					i++;
				x = i;
				while ((!isspace(radka[i]))
				       && (i < strlen(radka))) {
					extension[i - x] = radka[i];
					i++;
				}
				extension[i - x] = '\0';
				while ((isspace(radka[i]))
				       && (i < strlen(radka)))
					i++;
				x = i;
				if (i < strlen(radka)) {
					while (i < strlen(radka)) {
						content_type[i - x] =
						    radka[i];
						i++;
					}
				} else {
					fprintf(stderr,
						"invalid content_type definition on line %d\n",
						cur_conf_line);
					exit();
				}
				content_type[i - x - 1] = '\0';
				strcpy(pom->next->extension, extension);
				strcpy(pom->next->content_type,
				       content_type);
				pom->next->next = NULL;
			} else if (!strcasecmp(keyword, "virtualhost")) {
				/* parsing virtualhosts */
				pom2 = conf.virtual_hosts;
				while (pom2->next)
					pom2 = pom2->next;
				pom2->next =
				    (struct virtuals *)
				    malloc(sizeof(struct virtuals));
				/* fill in the three vals - parse according to spaces */
				i = 11;
				while (isspace(radka[i]))
					i++;
				x = i;
				while ((!isspace(radka[i]))
				       && (i < strlen(radka))) {
					name[i - x] = radka[i];
					i++;
				}
				name[i - x] = '\0';
				while ((isspace(radka[i]))
				       && (i < strlen(radka)))
					i++;
				x = i;
				if (i < strlen(radka)) {
					while ((!isspace(radka[i]))
					       && (i < strlen(radka))) {
						documentroot[i - x] =
						    radka[i];
						i++;
					}
					documentroot[i - x] = '\0';
					if (i < strlen(radka)) {
						while ((isspace(radka[i]))
						       && (i <
							   strlen(radka)))
							i++;
						x = i;
						while (i < strlen(radka)) {
							log[i - x] =
							    radka[i];
							i++;
						}
						log[i - x - 1] = '\0';
					} else {
						fprintf(stderr,
							"invalid virtualhost definition on line %d\n",
							cur_conf_line);
						exit();
					}
				} else {
					fprintf(stderr,
						"invalid content_type definition on line %d\n",
						cur_conf_line);
					exit();
				}
				strcpy(pom2->next->name, name);
				strcpy(pom2->next->documentroot,
				       documentroot);
				strcpy(pom2->next->log, log);
				pom2->next->next = NULL;
				/* open apropriate log file */
				if (!
				    (pom2->next->logfile =
				     fopen(log, "a"))) {
					fprintf(stderr,
						"! can't open %s!\n", log);
#ifdef HAVE_FCLOSEALL
					fcloseall();
#endif
					free_mem();
					exit();
				}
			} else
				fprintf(stderr,
					"warning: keyword %s on line %d not recognized\n",
					keyword, cur_conf_line);
		}
	}
	free(radka);
	free(confval);
	free(extension);
	free(content_type);
	free(name);
	free(documentroot);
	free(log);
}

char *find_content_type(char *ext)
{
	struct c_types *pom;

	pom = conf.content_types->next;
	while (pom) {
		if (!strcmp(ext, pom->extension))
			return (pom->content_type);
		pom = pom->next;
	}
	return ("text/html");
}

char *subst(char *a, char *b, char *c)
{
	int i = 0, x = 0, q;
	char *tmp;

	tmp = (char *) malloc(strlen(a) * strlen(b) * sizeof(char));
	while (i < strlen(a)) {
		if (a[i] != b[0]) {
			tmp[x] = a[i];
			x++;
			i++;
		} else {
			q = i;
			while (a[q] == b[q - i])
				q++;
			if (q - i == (strlen(b))) {	/* matches */
				strcat(tmp, c);
				i += strlen(b);
				x += strlen(c);
			} else {
				tmp[x] = a[i];
				x++;
				i++;
			}
		}
	}
	tmp[x] = '\0';
	return (tmp);
}

/* deletes duplicit occurences of char a in char* b */
char *deldupl(char a, char *b)
{
	int i = 0, x = 0;
	char *tmp;

	tmp = (char *) malloc(strlen(b) * sizeof(char));
	strcpy(tmp, "");
	while (i < strlen(b)) {
		if ((b[i] == a) && (b[i + 1] == a)) {
			tmp[i - x] = b[i];
			i++;
			while ((b[i] == a)) {
				i++;
				x++;
			}
		} else {
			tmp[i - x] = b[i];
			i++;
		}
	}
	tmp[i - x] = '\0';
	return tmp;
}

/* replaces %AB with corresponding character (0xAB) */

char *decode_hexa_octets(char *s1)
{
	int i, a, q;
	char hexa[3], *s2;

	s2 = (char *) malloc(sizeof(char) * strlen(s1) + 1);
	a = 0;
	i = 0;
	while (i < strlen(s1)) {
		if (s1[i] != '%') {
			s2[a] = s1[i];
			i++;
			a++;
		} else if (i < strlen(s1) - 1) {
			hexa[0] = s1[i + 1];
			hexa[1] = s1[i + 2];
			hexa[2] = '\0';
			q = strtol(hexa, NULL, 16);
			s2[a] = (char) q;
			a++;
			i += 3;
		}
	}
	s2[i] = '\0';
	return (s2);
}


 /* something very similar to realpath()
  * TODO: to be removed, and substituted with realpath()
  * on calling places
  */
int delddot(char *s, char *s2)
{
	int prev = 0, i = 0, q = 0, a = 0, w = 0, changed = 0;
	char *s1;

	s1 = (char *) malloc(sizeof(char) * MAXLINE);
	s1[0] = '\0';
	if (!strstr(s, "/../")) {
		strcpy(s2, s);
		return;
	}
	while ((i < strlen(s) - 4)
	       &&
	       (!((s[i] == '/') && (s[i + 1] == '.') && (s[i + 2] == '.')
		  && (s[i + 3] == '/')))) {
		if ((s[i] == '/') && (i != 0)) {
			for (a = prev; a < i; a++) {
				s1[a] = s[a];
			}
			prev = a;
			s1[a] = '\0';
		}
		i++;
	}
	if (i < strlen(s) - 4) {
		changed = 1;
		i = i + 3;
		q = strlen(s1);
		w = i;
		while (i < strlen(s)) {
			s1[q + i - w] = s[i];
			i++;
		}
		s1[q + i - w] = '\0';
		delddot(s1, s2);	/* recursively on the same string, d0h ;) */
	}
	free(s1);
}

int get_number_of_month(char *tmp2)
{
	if (!strcasecmp(tmp2, "Jan"))
		return 1;
	if (!strcasecmp(tmp2, "Feb"))
		return 2;
	if (!strcasecmp(tmp2, "Mar"))
		return 3;
	if (!strcasecmp(tmp2, "Apr"))
		return 4;
	if (!strcasecmp(tmp2, "May"))
		return 5;
	if (!strcasecmp(tmp2, "Jun"))
		return 6;
	if (!strcasecmp(tmp2, "Jul"))
		return 7;
	if (!strcasecmp(tmp2, "Aug"))
		return 8;
	if (!strcasecmp(tmp2, "Sep"))
		return 9;
	if (!strcasecmp(tmp2, "Oct"))
		return 10;
	if (!strcasecmp(tmp2, "Nov"))
		return 11;
	if (!strcasecmp(tmp2, "Dec"))
		return 12;
}


int compare_rfc822_dates(char *date1, char *date2)
 /* compares two dates. 
  * 
  * if (d1 < d2) return (-1); else if (d1 == d2) return(0); else return(1);
  * 
  */
{
	int i, bb;
	int long yr1 = 0, yr2 = 0;
	int mon1 = 0, mon2 = 0;
	int day1 = 0, day2 = 0;
	int hms1 = 0, hms2 = 0;
	int offs1 = 0, offs2 = 0;
	int num1, num2;
	char *tmp, tmp2[2048], op[2], b[3];

	/* parse years into yr1 & yr2 */

	tmp = strpbrk(date1, " ");
	tmp++;
	tmp = strpbrk(tmp, " ");
	tmp++;
	tmp = strpbrk(tmp, " ");
	tmp++;

	for (i = 0; i < 4; i++)
		tmp2[i] = tmp[i];
	tmp2[i] = '\0';
	yr1 = atoi(tmp2);

	/* yr2 */
	tmp = strpbrk(date2, " ");
	tmp++;
	tmp = strpbrk(tmp, " ");
	tmp++;
	tmp = strpbrk(tmp, " ");
	tmp++;
	for (i = 0; i < 4; i++)
		tmp2[i] = tmp[i];
	tmp2[i] = '\0';
	yr2 = atoi(tmp2);

	/* parse days to d1 & d2 */

	tmp = strpbrk(date1, " ");
	tmp++;
	for (i = 0; i < 2; i++)
		tmp2[i] = tmp[i];
	tmp2[i] = '\0';
	day1 = atoi(tmp2);

	tmp = strpbrk(date2, " ");
	tmp++;
	for (i = 0; i < 2; i++)
		tmp2[i] = tmp[i];
	tmp2[i] = '\0';
	day2 = atoi(tmp2);

	/* parse hms */

	tmp = strpbrk(date1, " ");
	tmp++;
	for (i = 0; i < 3; i++) {
		tmp = strpbrk(tmp, " ");
		tmp++;
	}
	strcpy(tmp2, tmp);
	tmp2[8] = '\0';
	tmp = subst(tmp2, ":", "");
	hms1 = atoi(tmp);

	tmp = strpbrk(date2, " ");
	tmp++;
	for (i = 0; i < 3; i++) {
		tmp = strpbrk(tmp, " ");
		tmp++;
	}
	strcpy(tmp2, tmp);
	tmp2[8] = '\0';
	tmp = subst(tmp2, ":", "");
	hms2 = atoi(tmp);

	/* count 'offset' */

	tmp = strpbrk(date1, " ");
	tmp++;
	for (i = 0; i < 4; i++) {
		tmp = strpbrk(tmp, " ");
		tmp++;
	}
	strcpy(tmp2, tmp);
	if (!strncmp(tmp2, "GMT", 3))
		strcpy(tmp2, "+0000");
	b[0] = tmp2[1];
	b[1] = tmp2[2];
	b[3] = '\0';
	bb = atoi(b);
	if (tmp2[0] == '+')
		hms1 = hms1 + bb * 100000;

	tmp = strpbrk(date2, " ");
	tmp++;

	for (i = 0; i < 4; i++) {
		tmp = strpbrk(tmp, " ");
		tmp++;
	}
	strcpy(tmp2, tmp);
	if (!strncmp(tmp2, "GMT", 3))
		strcpy(tmp2, "+0000");
	b[0] = tmp2[1];
	b[1] = tmp2[2];
	b[3] = '\0';
	bb = atoi(b);
	if (tmp2[0] == '+')
		hms2 = hms2 + bb * 100000;

	/* count month # */

	tmp = strpbrk(date1, " ");
	tmp++;
	for (i = 0; i < 1; i++) {
		tmp = strpbrk(tmp, " ");
		tmp++;
	}
	strcpy(tmp2, tmp);
	tmp2[3] = '\0';
	mon1 = get_number_of_month(tmp2);

	tmp = strpbrk(date2, " ");
	tmp++;
	for (i = 0; i < 1; i++) {
		tmp = strpbrk(tmp, " ");
		tmp++;
	}
	strcpy(tmp2, tmp);
	tmp2[3] = '\0';
	mon2 = get_number_of_month(tmp2);

	if (yr1 < yr2)
		return -1;
	else if (yr1 > yr2)
		return 1;
	else if (mon1 < mon2)
		return -1;
	else if (mon1 > mon2)
		return 1;
	else if (day1 < day2)
		return -1;
	else if (day1 > day2)
		return 1;
	else if (hms1 < hms2)
		return -1;
	else if (hms1 > hms2)
		return 1;
	else if (hms1 == hms2)
		return 0;
}
